/*{{{  file header */

/*

    Meta-Environment - An environment for language prototyping.
    Copyright (C) 2000  Stichting Mathematisch Centrum, Amsterdam,
                        The Netherlands.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

*/
/*
  $Id: focus.c,v 1.4 2000/05/15 15:42:24 markvdb Exp $  
 */

/*
 * The version of gse is written by Mark van den Brand.
 */

/*}}}  */
/*{{{  includes */

#include <stdlib.h>

#include "se.h"   
#include "area.h"
#include "editor.h"
#include "focus.h"
#include "path.h"
#include "sort.h"
#include "tree.h"

/*}}}  */

/*{{{  ATerm newFocus(ATerm path, ATerm sort, ATerm area) */

ATerm newFocus(ATerm path, ATerm sort, ATerm area)
{
  return ATmake("focus(<term>,<term>,<term>)", path, sort, area);
}

/*}}}  */
/*{{{  ATerm setDirty(ATerm focus) */

ATerm setDirty(ATerm focus)
{
  return ATsetAnnotation(focus, ATparse("focusstatus"),ATparse("dirty"));
}

/*}}}  */
/*{{{  ATerm setClean(ATerm focus) */

ATerm setClean(ATerm focus)
{
  return ATsetAnnotation(focus, ATparse("focusstatus"),ATparse("clean"));
}

/*}}}  */
/*{{{  ATerm newFocusList(ATerm focus) */

ATerm newFocusList(ATerm focus)
{
  return (ATerm)ATmakeList1(focus);
}

/*}}}  */
/*{{{  ATerm getFocusPath(ATerm focus) */

ATerm getFocusPath(ATerm focus)
{
  ATerm path;

  if (ATmatch(focus, "focus(<term>,<term>,<term>)", &path, NULL, NULL)) {
    return path;
  }
  else {
    ATerror("Illegal focus: %t\n", focus);
    return NULL;
  }       
}

/*}}}  */
/*{{{  ATerm setFocusSort(ATerm focus, ATerm newSort) */

ATerm setFocusSort(ATerm focus, ATerm newSort)
{
  ATerm path, oldSort, area;

  if (ATmatch(focus, "focus(<term>,<term>,<term>)", &path, &oldSort, &area)) {
    return ATmake("focus(<term>,<term>,<term>)", path, newSort, area);
  }
  else {
    ATerror("Illegal focus: %t\n", focus);
    return focus;
  }       
}

/*}}}  */
/*{{{  ATerm getFocusSort(ATerm focus) */

ATerm getFocusSort(ATerm focus)
{
  ATerm sort;

  if (ATmatch(focus, "focus(<term>,<term>,<term>)", NULL, &sort, NULL)) {
    return sort;
  }
  else {
    ATerror("Illegal focus: %t\n", focus);
    return NULL;
  }       
}

/*}}}  */
/*{{{  ATerm getFocusArea(ATerm focus) */

ATerm getFocusArea(ATerm focus)
{
  ATerm area;

  if (ATmatch(focus, "focus(<term>,<term>,<term>)", NULL, NULL, &area)) {
    return area;
  }
  else {
    ATerror("Illegal focus: %t\n", focus);
    return NULL;
  }       
}

/*}}}  */
/*{{{  static ATerm setFocusArea(ATerm focus, ATerm newArea) */

static ATerm setFocusArea(ATerm focus, ATerm newArea)
{
  ATerm path, sort, oldArea;

  if (ATmatch(focus, "focus(<term>,<term>,<term>)", &path, &sort, &oldArea)) {
    return ATmake("focus(<term>,<term>,<term>)", path, sort, newArea);
  }
  else {
    ATerror("Illegal focus: %t\n", focus);
    return focus;
  }       
}

/*}}}  */
/*{{{  ATerm createFocusGivenPath(ATerm tree, ATerm path) */

ATerm createFocusGivenPath(ATerm tree, ATerm path)
{
  ATerm subTree = subtreeOfTreeGivenPath(tree, path);
  ATerm sort = getReadableSort(getTreeSort(subTree));  
  ATerm area = createArea(path, tree);

  return ATmake("focus(<term>,<term>,<term>)", path, sort, area); 
}

/*}}}  */
/*{{{  ATerm createFocusGivenLocation(ATerm tree, int i)
 */

ATerm createFocusGivenLocation(ATerm tree, int i)
{
  ATerm newFocus;
  ATerm path = getPathInTree(tree, i, newPath());
  newFocus = createFocusGivenPath(tree, path);
  return setClean(newFocus);
}

/*}}}  */
/*{{{  ATbool inFocusArea(ATerm focus, int loc) */

ATbool inFocusArea(ATerm focus, int loc)
{
  ATerm area = getFocusArea(focus);
  return inArea(area, loc);
}

/*}}}  */
/*{{{  ATbool afterFocusArea(ATerm focus, int loc) */

ATbool afterFocusArea(ATerm focus, int loc)
{
  ATerm area = getFocusArea(focus);
  return afterArea(area, loc);
}

/*}}}  */
/*{{{  ATbool isChild(ATerm focusOne, ATerm focusTwo) */

ATbool isChild(ATerm focusOne, ATerm focusTwo)
{
  ATerm pathOne = getFocusPath(focusOne);
  ATerm pathTwo = getFocusPath(focusTwo);
  return isChildPath(pathOne, pathTwo);
}

/*}}}  */
/*{{{  ATermList removeChildren(ATerm focus, ATermList focuses) */

ATermList removeChildren(ATerm focus, ATermList focuses)
{
  if(ATisEmpty(focuses)) {
    return ATmakeList1(focus);
  }
  else {
    ATerm firstFocus = ATgetFirst(focuses);
    focuses = ATgetNext(focuses);
    if(isChild(firstFocus, focus)) {
      return removeChildren(focus, focuses);
    }
    else {
      ATermList newFocuses = removeChildren(focus, focuses);
      return ATinsert(newFocuses, firstFocus);
    }
  }
}

/*}}}  */
/*{{{  ATermList updateFocus(ATermList focuses, int location, int length) */

ATermList updateFocus(ATermList focuses, int location, int length)
{
  ATermList newFocuses;

  if(ATisEmpty(focuses)) {
    return focuses;
  }
  else {
    ATerm firstFocus = ATgetFirst(focuses);
    focuses = ATgetNext(focuses);
    if(afterFocusArea(firstFocus, location)) {
      newFocuses = updateFocus(focuses, location, length);
      return ATinsert(newFocuses, firstFocus);
    }
    else {
      ATerm area = getFocusArea(firstFocus);
      if(inFocusArea(firstFocus, location)) {
        int areaLength = getIntLengthOfArea(area);
        ATerm newArea = setIntLengthOfArea(area, areaLength+length);
        ATerm newFocus = setFocusArea(firstFocus, newArea);
        newFocuses = updateFocus(focuses, location, length);
        return ATinsert(newFocuses, newFocus);
      }
      else {
        int areaStart = getIntStartOfArea(area);
        ATerm newArea = setIntStartOfArea(area, areaStart+length);
        ATerm newFocus = setFocusArea(firstFocus, newArea);
        newFocuses = updateFocus(focuses, location, length);
        return ATinsert(newFocuses, newFocus);
      }
    }
  }
}

/*}}}  */
/*{{{  ATerm applyMoveToFocus(ATerm editor, ATerm move) */

ATerm applyMoveToFocus(ATerm editor, ATerm move)
{
  ATermList focuses = (ATermList)getFocuses(editor), newFocuses;
  ATerm firstFocus = ATgetFirst(focuses), newFocus;
  ATerm path = getFocusPath(firstFocus), newPath;
  ATerm tree = getTreeInEditor(editor), subTree;

  if(ATisEqual(move, ATparse(MOVE_UP))) {
    newPath = goUp(path);
    newFocus = createFocusGivenPath(tree, newPath);
    focuses = ATgetNext(focuses);
    newFocuses = ATinsert(focuses, newFocus);
    return putFocuses(editor, (ATerm)newFocuses);
  }                                                 
  else if(ATisEqual(move, ATparse(MOVE_DOWN))) {
    newPath = goDown(path);
    subTree = subtreeOfTreeGivenPath(tree, newPath);
    if(subTree) {
      if(asfix_is_appl(subTree) ||
         asfix_is_list(subTree) ||
         asfix_is_lex(subTree) ||
         asfix_is_var(subTree)) {
        newFocus = createFocusGivenPath(tree, newPath);
        focuses = ATgetNext(focuses);
        newFocuses = ATinsert(focuses, newFocus);
        return putFocuses(editor, (ATerm)newFocuses);
      }
      else {
        newFocus = createFocusGivenPath(tree, newPath);
        newFocuses = ATinsert(focuses, newFocus);
        editor = putFocuses(editor, (ATerm)newFocuses);
        return applyMoveToFocus(editor, ATparse(MOVE_RIGHT));
      }
    }
    return editor;
  }
  else if(ATisEqual(move, ATparse(MOVE_RIGHT))) {
    newPath = goRight(path);
    subTree = subtreeOfTreeGivenPath(tree, newPath);
    if(subTree) {                            
      newFocus = createFocusGivenPath(tree, newPath);
      focuses = ATgetNext(focuses);
      if(ATgetType(subTree) == AT_APPL) {
        if(!(asfix_is_appl(subTree) ||
             asfix_is_list(subTree) ||
             asfix_is_lex(subTree) ||
             asfix_is_var(subTree))) {
          newFocuses = ATinsert(focuses, newFocus);
          editor = putFocuses(editor, (ATerm)newFocuses);
          return applyMoveToFocus(editor, move);
        }
      }
      newFocuses = ATinsert(focuses, newFocus);
      return putFocuses(editor, (ATerm)newFocuses);
    }
    return applyMoveToFocus(editor, ATparse(MOVE_UP));
    /*return editor;*/
  }
  else if(ATisEqual(move, ATparse(MOVE_LEFT))) {
    newPath = goLeft(path);
    if(!ATisEqual(newPath, path)) {
      subTree = subtreeOfTreeGivenPath(tree, newPath);
      if(subTree) {                            
        newFocus = createFocusGivenPath(tree, newPath);
        focuses = ATgetNext(focuses);
        if(ATgetType(subTree) == AT_APPL) {
          if(!(asfix_is_appl(subTree) ||
               asfix_is_list(subTree) ||
               asfix_is_lex(subTree) ||
               asfix_is_var(subTree))) {
            newFocuses = ATinsert(focuses, newFocus);
            editor = putFocuses(editor, (ATerm)newFocuses);
            return applyMoveToFocus(editor, move);
          }
        }
        newFocuses = ATinsert(focuses, newFocus);
        return putFocuses(editor, (ATerm)newFocuses);
      }
      return applyMoveToFocus(editor, ATparse(MOVE_UP));
    }
    else {
      return applyMoveToFocus(editor, ATparse(MOVE_UP));
    }
  }
  else {
    ATerror("Illegal move action\n");
    return editor;
  }
}

/*}}}  */

